/*-----------------------------------------------------------------------------
File:          json_command_interpreter.cpp
Date:          210705
Last modified: 210729
By Audun Brabrand & Tord Brabrand
-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <logging/log.h>
#include <dk_buttons_and_leds.h>
#include "json_command_interpreter.h"


LOG_MODULE_REGISTER(json_command_interpreter, CONFIG_CLOUD_CLIENT_LOG_LEVEL);

/*-----------------------------------------------------------------------------
    Definition of an <attribute : value> - pair:
*/
#define MAX_TOKEN_SIZE 40
#define MAX_INT        32767

// Type of value:
#define NUM_TYPE       0
#define TEXT_TYPE      1

typedef struct attrib_s
{
	unsigned char *name;  // Name of this attribute
	union
	{
		int  num_value;
		unsigned char text_value[MAX_TOKEN_SIZE];
	} value;
	int type;                   // Type of value
	bool value_updated;         // Flag telling if the value of this attribute has been updated
} ATTRIB_T;

/*-----------------------------------------------------------------------------
    Attribute table:
*/
#define a_TEMP_REPORTING          0
#define a_LED1		          1
#define a_LED2		          2
#define a_LED3		          3
#define a_LED4		          4
#define a_GET		          5

#define LAST_ATTRIBUTE        a_GET
#define NO_OF_ATTRIBUTES      (LAST_ATTRIBUTE+1)
#define UNKNOWN_ATTRIB        (NO_OF_ATTRIBUTES+1)

// NB: All attribute names must be uppercase!!
ATTRIB_T attribute[NO_OF_ATTRIBUTES] =
{
	{ "TEMP_REPORTING", {0}, NUM_TYPE,  TRUE },
	{ "LED1",           {0}, NUM_TYPE,  TRUE },
	{ "LED2",           {0}, NUM_TYPE,  TRUE },
	{ "LED3",           {0}, NUM_TYPE,  TRUE },
	{ "LED4",           {0}, NUM_TYPE,  TRUE },
	{ "GET",            {0}, NUM_TYPE,  TRUE }  /* Used as a command, - must be the last "attribute" */
};

/*---------------------------------------------------------------------------*/
int my_isdigit(unsigned char in)
{
	if ((in >= '0') && (in <= '9'))
		return(TRUE);
	else
		return(FALSE);
}
/*---------------------------------------------------------------------------*/
unsigned char my_tolower(unsigned char in)
{
	if ((in >= 'A') && (in <= 'Z'))
		return((in + ('a'-'A')));
	else
		return(in);
}
/*---------------------------------------------------------------------------*/
unsigned char my_toupper(unsigned char in)
{
	if ((in >= 'a') && (in <= 'z'))
		return((in - ('a' - 'A')));
	else
		return(in);
}
/*---------------------------------------------------------------------------*/
int my_strcmp(unsigned char *in_p, unsigned char *key_p)
{
	while (1)
	{
		if (my_toupper(*in_p) != *key_p)
			return(1);

		if (*in_p == 0)
			return(0);

		in_p++;
		key_p++;
	}
}
/*---------------------------------------------------------------------------*/
int my_strcopy(unsigned char *dest_p, unsigned char *source_p)
{
	int length;

	length = 0;
	while (*source_p != 0)
	{
		*(dest_p++) = *(source_p++);
		length++;
	}
	*(dest_p++) = *(source_p++);  // Copy terminating 0
	length++;   // Length includes terminating 0
	return(length);
}
/*---------------------------------------------------------------------------*/
#define MAX_JSON_TX_LENGTH  256
unsigned char json_tx_buffer[MAX_JSON_TX_LENGTH];
unsigned char json_temp_buffer[MAX_JSON_TX_LENGTH];
bool get_command_received = FALSE;

/*----------------------------------------*/
// Make json string to deliver temperature to nRF Cloud:
#ifdef FLOAT_TEMP
void MakeJsonTemperatureString(float temperature)  // TODO: Make sure it doesn't write outside json_temp_buffer!!!
{
   unsigned char *tx_p;

    tx_p = json_temp_buffer;
    tx_p += sprintf((char*)tx_p, "{\"appId\":\"TEMP\", \"data\": %.1f, \"messageType\": \"DATA\"}", temperature);
}
#else
void MakeJsonTemperatureString(int temperature)  // TODO: Make sure it doesn't write outside json_temp_buffer!!!
{
   unsigned char *tx_p;

    tx_p = json_temp_buffer;
    tx_p += sprintf((char*)tx_p, "{\"appId\":\"TEMP\", \"data\": %d, \"messageType\": \"DATA\"}", temperature);
}
#endif
/*----------------------------------------*/
// Make json string of specified attribute:
void MakeJsonStatusString(int attrib)  // TODO: Make sure it doesn't write outside json_tx_buffer!!!
{
    unsigned char *tx_p;

    tx_p = json_tx_buffer;
    *(tx_p++) = '{';

    if (attrib < NO_OF_ATTRIBUTES)
    {
        *(tx_p++) = '"';
        tx_p += (my_strcopy(tx_p, attribute[attrib].name)) - 1;
        tx_p += (my_strcopy(tx_p, "\" : ")) - 1;
        if (attribute[attrib].type == TEXT_TYPE)
        {
            *(tx_p++) = '"';
            tx_p += (my_strcopy(tx_p, attribute[attrib].value.text_value)) - 1;
            *(tx_p++) = '"';
        }
        else
            tx_p += sprintf((char*)tx_p, "%d", attribute[attrib].value.num_value);
    }
    *(tx_p++) = '}';
    *(tx_p++) = 0;
}
/*----------------------------------------*/
// Make json string of all attributes:
void MakeJsonStatusStringAll(void)  // TODO: Make sure it doesn't write outside json_tx_buffer!!!
{
    int i;
    unsigned char *tx_p;

    tx_p = json_tx_buffer;
    *(tx_p++) = '{';

    for (i=0; i<NO_OF_ATTRIBUTES; i++)
    {
        if (i != 0)
        {
            *(tx_p++) = ',';
            *(tx_p++) = ' ';
        }
	*(tx_p++) = '"';
        tx_p += (my_strcopy(tx_p, attribute[i].name)) - 1;
        tx_p += (my_strcopy(tx_p, "\" : ")) - 1;
        if (attribute[i].type == TEXT_TYPE)
        {
            *(tx_p++) = '"';
            tx_p += (my_strcopy(tx_p, attribute[i].value.text_value)) - 1;
            *(tx_p++) = '"';
        }
        else
            tx_p += sprintf((char*)tx_p, "%d", attribute[i].value.num_value);
    }
    *(tx_p++) = '}';
    *(tx_p++) = 0;
}
/*---------------------------------------------------------------------------*/
int temp_reporting = 0;
/*---------------------------------------------------------------------------*/
#define GET_TEMP_REPORTING   a_TEMP_REPORTING
#define GET_LED1             a_LED1
#define GET_LED2             a_LED2
#define GET_LED3             a_LED3
#define GET_LED4             a_LED4
#define GET_ALL              1000

void HandleNewValue(int attrib, int val)  // val: 0 => off, else => on
{
      switch (attrib)
      {
      case a_TEMP_REPORTING:
          temp_reporting = val;
          break;

      case a_LED1:
          dk_set_led(DK_LED1, (val? 0:1));
          break;

      case a_LED2:
          dk_set_led(DK_LED2, (val? 0:1));
          break;

      case a_LED3:
          dk_set_led(DK_LED3, (val? 0:1));
          break;

      case a_LED4:
          dk_set_led(DK_LED4, (val? 0:1));
          break;

      case a_GET:
          if (val == GET_ALL)
              MakeJsonStatusStringAll();
          else
              MakeJsonStatusString(val);
          get_command_received = TRUE;
          break;

      default:
          break;
      }
}
/*---------------------------------------------------------------------------*/
void HandleUpdatedAttibuteValues(void)
{
    int att;

    for (att=0; att<NO_OF_ATTRIBUTES; att++)  // Check all atttributes:
    {
        if (attribute[att].value_updated)     // If attribute has been updated...
        {
            if (att == a_GET)  // Special handling of attribute GET (used as command):
            {
                  if (attribute[att].type == TEXT_TYPE)
                  {
                      if (my_strcmp(attribute[att].value.text_value, "ALL") == 0)
                            HandleNewValue(att, GET_ALL);
                      else if (my_strcmp(attribute[att].value.text_value, "LED1") == 0)
                            HandleNewValue(att, GET_LED1);
                      else if (my_strcmp(attribute[att].value.text_value, "LED2") == 0)
                            HandleNewValue(att, GET_LED2);
                      else if (my_strcmp(attribute[att].value.text_value, "LED3") == 0)
                            HandleNewValue(att, GET_LED3);
                      else if (my_strcmp(attribute[att].value.text_value, "LED4") == 0)
                            HandleNewValue(att, GET_LED4);
                      else if (my_strcmp(attribute[att].value.text_value, "TEMP_REPORTING") == 0)
                            HandleNewValue(att, GET_TEMP_REPORTING);
                  }
                  else if (attribute[att].type == NUM_TYPE)
                  {
                      HandleNewValue(att, attribute[att].value.num_value);
                  }
            }
            else  // Not a GET command, just update the attribute value:
            {
                  if (attribute[att].type == TEXT_TYPE)
                  {
                    if (my_strcmp(attribute[att].value.text_value, "ON") == 0)
                          HandleNewValue(att, 1);
                    else if (my_strcmp(attribute[att].value.text_value, "OFF") == 0)
                          HandleNewValue(att, 0);
                  }
                  else if (attribute[att].type == NUM_TYPE)
                  {
                    if (attribute[att].value.num_value == 0)
                          HandleNewValue(att, 0);
                    else
                          HandleNewValue(att, 1);
                  }
            }
            attribute[att].value_updated = FALSE;
        }
    }
}
/*---------------------------------------------------------------------------*/
void InitJsonCommandInterpreter(void)
{
      int err;

      // Initialize LED handler:
      err = dk_leds_init();
      if (err)
          LOG_ERR("dk_leds_init, error: %d", err);
      else
          LOG_INF("LED handler initialized.\n");
      
      HandleUpdatedAttibuteValues();  // Make sure status of LEDs etc. matches attribute values in attribute table
}
/*-----------------------------------------------------------------------------
Global variables for json parser:
*/
unsigned char *in_p;
unsigned char token[MAX_TOKEN_SIZE];
int current_attribute = UNKNOWN_ATTRIB;  // Index into attribute table

/*-----------------------------------------------------------------------------
Tags returned by GetToken:
*/
#define t_END_OF_STRING       0
#define t_PAR_START           1
#define t_PAR_END             2
#define t_COLON               3
#define t_APOST               4
#define t_PLUS                5
#define t_MINUS               6
#define t_COMMA               7
#define t_WORD                8
#define t_NUM                 9
// Error tags:
#define t_TOO_LONG            100
#define t_UNKNOWN             101  // Must be the last symbol!!

/*---------------------------------------------------------------------------*/
/*
int CheckKeyword(void)
{
	int i;

	for (i = 0; keyword[i].tag != t_UNKNOWN; i++)
	{
		if (my_strcmp(token, keyword[i].token) == 0)
		{
			return(keyword[i].tag);
		}
	}
	return(t_UNKNOWN);
}
*/
/*---------------------------------------------------------------------------*/
int ConvertNumber(unsigned char *num_string)
{
	int value, i, first, n;

	i = 0;
	// Search for first non-zero digit:
	while (num_string[i] == '0')
		i++;
	first = i;  // Index to first non-zero digit
	// Find end of string:
	while (num_string[i] != 0)
		i++;
	i--;
	// i is now index to last digit:
	value = 0;
	n = 1;
	while (i >= first)  // For all digits in string...
	{
		value += (num_string[i] - '0') * n;
		n *= 10;
		i--;
		if (value > MAX_INT)
		{
			LOG_ERR("Value is too big!\n");
			return (0);
		}
	}
	return (value);
}
/*---------------------------------------------------------------------------*/
void PrintAttribTable(void)
{
	int i;

	for (i = 0; i < NO_OF_ATTRIBUTES; i++)
	{
		if (attribute[i].type == NUM_TYPE)
			LOG_INF("%d: %s = %d", i, attribute[i].name, attribute[i].value.num_value);
		else
			LOG_INF("%d: %s = %s", i, attribute[i].name, attribute[i].value.text_value);

		if (attribute[i].value_updated)
			LOG_INF(" (updated)\n");
		else
			LOG_INF("\n");
	}
}
/*---------------------------------------------------------------------------*/
// in_p must be set to point to start of input string before first call to this function!
// input string must be 0-terminated!
int GetToken(void)
{
	unsigned char *token_p;
	int length;

	// Skip whitespace and CRLF:
	while ((*in_p == ' ') || (*in_p == '\t') || (*in_p == '\r') || (*in_p == '\n'))
		in_p++;

	// Check next char:
	switch (*in_p)
	{
	case '{':
		in_p++;
		return(t_PAR_START);

	case '}':
		in_p++;
		return(t_PAR_END);

	case ':':
		in_p++;
		return(t_COLON);

	case '"':
		in_p++;
		return(t_APOST);

	case '+':
		in_p++;
		return(t_PLUS);

	case '-':
		in_p++;
		return(t_MINUS);

	case ',':
		in_p++;
		return(t_COMMA);

	case 0:
		return(t_END_OF_STRING);

	default:
		break;
	}
	if (isalpha(*in_p))  // Handle a word:
	{
		token_p = token;
		length = 0;

		while ((isalnum(*in_p)) || (*in_p == '_'))
		{
			if (length < (MAX_TOKEN_SIZE-1))
			{
				*(token_p++) = *(in_p++);  // Copy token to "token"
				length++;
			}
			else
			{
				*token_p = 0;  // Too long input word, add 0 and end
				return(t_TOO_LONG);
			}
		}
		*(token_p++) = 0;  // Add terminating 0
		length++;
		return(t_WORD);
	}
	else if (isdigit(*in_p))  // Handle a number:
	{
		token_p = token;
		length = 0;

		while (isdigit(*in_p))
		{
			if (length < (MAX_TOKEN_SIZE - 1))
			{
				*(token_p++) = *(in_p++);  // Copy token to "token"
				length++;
			}
			else
			{
				*token_p = 0;  // Too long input word or number, add 0 and end
				return(t_TOO_LONG);
			}
		}
		*(token_p++) = 0;  // Add terminating 0
		length++;
		return(t_NUM);
	}
	return(t_UNKNOWN);
}
/*---------------------------------------------------------------------------*/
// Prototypes:
void GetAttribute(void);
/*---------------------------------------------------------------------------*/
void GetValue(void)  // Get value for current_attribute:
{
	switch (GetToken())
	{
	case t_PLUS:
		if (GetToken() == t_NUM)
		{
			attribute[current_attribute].value.num_value = ConvertNumber(token);
			attribute[current_attribute].type = NUM_TYPE;
			attribute[current_attribute].value_updated = TRUE;
		}
		else
		{
			LOG_ERR("Expecting a number!\n");
			return;
		}
		break;

	case t_MINUS:
		if (GetToken() == t_NUM)
		{
			attribute[current_attribute].value.num_value = -(ConvertNumber(token));
			attribute[current_attribute].type = NUM_TYPE;
			attribute[current_attribute].value_updated = TRUE;
		}
		else
		{
			LOG_ERR("Expecting a number!\n");
			return;
		}
		break;

	case t_NUM:
		attribute[current_attribute].value.num_value = ConvertNumber(token);
		attribute[current_attribute].type = NUM_TYPE;
		attribute[current_attribute].value_updated = TRUE;
		break;

	case t_APOST:  // Start of textual value
		if (GetToken() == t_WORD)
		{
			my_strcopy(attribute[current_attribute].value.text_value, token);
			attribute[current_attribute].type = TEXT_TYPE;
			attribute[current_attribute].value_updated = TRUE;

			if (GetToken() == t_APOST)  // End-apostroph
			{
				;
			}
			else
			{
				LOG_ERR("Missing '\"'!\n");
				return;
			}
		}
		else
		{
			LOG_ERR("Expecting a word!\n");
			return;
		}
		break;

	default:
		LOG_ERR("Unknown value!\n");
		return;
	}
	// Value found, continue parsing:
	switch (GetToken())
	{
	case t_PAR_END:  // End of json-string - finished!
		break;

	case t_COMMA:
		// Get another <attribute : value> - pair:
		GetAttribute();
		break;

	default:
		LOG_ERR("Unknown value!\n");
		break;
	}
}
/*---------------------------------------------------------------------------*/
void GetAttribute(void)  // Get attribute
{
	int i;

	// Get first '"':
	if (GetToken() == t_APOST)
	{
		if (GetToken() == t_WORD)
		{
			current_attribute = UNKNOWN_ATTRIB;

			// Check if this word is a known attribute:
			for (i = 0; i < NO_OF_ATTRIBUTES; i++)
			{
				if (my_strcmp(token, attribute[i].name) == 0)
				{
					current_attribute = i;
					break;
				}
			}
			if (current_attribute != UNKNOWN_ATTRIB)
			{
				// Get second '"':
				if (GetToken() == t_APOST)
				{
					// Get ':':
					if (GetToken() == t_COLON)
					{
						// Get value for the attribute:
						GetValue();
					}
					else
						LOG_ERR("Missing ':'!\n");
				}
				else
					LOG_ERR("Missing '\"'!\n");
			}
			else
				LOG_ERR("Unknown attribute \"%s\"!\n", token);
		}
		else
			LOG_ERR("Unknown attribute \"%s\"!\n", token);
	}
	else
		LOG_ERR("Missing '\"'!\n");
}
/*---------------------------------------------------------------------------*/
void GetParStart(void)  // Get {
{
	if (GetToken() == t_PAR_START)
	{
		GetAttribute();
	}
	else
		LOG_ERR("Missing '{'!\n");
}
/*---------------------------------------------------------------------------*/
void ParseJsonString(unsigned char *js_p)
{
	in_p = js_p;
	GetParStart();
}
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/


