/*
 * Copyright (c) 2020 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

#include <zephyr.h>
#include <stdio.h>
#include <modem/lte_lc.h>
#include <net/cloud.h>
#include <net/socket.h>
#include <dk_buttons_and_leds.h>
#include <logging/log.h>

LOG_MODULE_REGISTER(cloud_client, CONFIG_CLOUD_CLIENT_LOG_LEVEL);

//========================================================
/*  ----- Extra includes: ---------------------------  */
#include "json_command_interpreter.h"
#include <drivers/sensor.h>

/*  ----- Buffer for json messages from nRF Cloud ---  */
unsigned char rx_buffer[256];

/*  ----- Semaphore for json messages reception -----  */
static K_SEM_DEFINE(json_received, 0, 1);

/*  ----- Prototype function ------------------------- */ 
void SendTemperatureToCloud(void);

/*  ----- Tempaerature thread config ----------------  */
#define STACKSIZE                 1024
#define PRIORITY                  7
#define TEMP_REPORTING_INTERVAL   5000


void send_temperature_thread(void)
{
    const struct device *dev;
    struct sensor_value data;
    struct k_spinlock lock;
    k_spinlock_key_t key;
    double value = 0;
    int err;

    // Adding sensor simulator to project (the nRF9160-DK board doens't have a temperature sensor, only a SENSOR_SIM simulator):
    dev = device_get_binding(CONFIG_SENSOR_SIM_DEV_NAME);
    __ASSERT(dev, "Could not get device %s\n", CONFIG_SENSOR_SIM_DEV_NAME);

    while(1)
    {
        k_msleep(TEMP_REPORTING_INTERVAL);
        LOG_DBG("Temp reporting interval expired.\n");

        if (temp_reporting)
        {
            err = sensor_sample_fetch_chan(dev, SENSOR_CHAN_AMBIENT_TEMP);
            if (err)
                LOG_ERR("Failed to fetch data from %s, error: %d", CONFIG_SENSOR_SIM_DEV_NAME, err);

            err = sensor_channel_get(dev, SENSOR_CHAN_AMBIENT_TEMP, &data);
            if (err)
                LOG_ERR("Failed to fetch data from %s, error: %d", CONFIG_SENSOR_SIM_DEV_NAME, err);
            else
            {   // Make sure no other thread can change the data value while we are converting to "double"
                key = k_spin_lock(&lock);
                value = sensor_value_to_double(&data);
                k_spin_unlock(&lock, key);
            }
            MakeJsonTemperatureString((float)value);
            LOG_INF("Sending temperature to cloud: %.1f\n", (float)value);
            SendTemperatureToCloud();
        }
    }
}

K_THREAD_DEFINE(send_temperature_thread_id, STACKSIZE, send_temperature_thread, NULL, NULL, NULL,
		PRIORITY, 0, 0);

//========================================================
static struct cloud_backend *cloud_backend;
static struct k_delayed_work cloud_update_work;
static struct k_delayed_work connect_work;

static K_SEM_DEFINE(lte_connected, 0, 1);

/* Flag to signify if the cloud client is connected or not connected to cloud,
 * used to abort/allow cloud publications.
 */
static bool cloud_connected;

static void connect_work_fn(struct k_work *work)
{
	int err;

	err = cloud_connect(cloud_backend);
	if (err) {
		LOG_ERR("cloud_connect, error: %d", err);
	}

	LOG_INF("Next connection retry in %d seconds",
	       CONFIG_CLOUD_CONNECTION_RETRY_TIMEOUT_SECONDS);

	k_delayed_work_submit(&connect_work,
		K_SECONDS(CONFIG_CLOUD_CONNECTION_RETRY_TIMEOUT_SECONDS));
}

static void cloud_update_work_fn(struct k_work *work)
{
	int err;

	if (!cloud_connected) {
		LOG_INF("Not connected to cloud, abort cloud publication");
		return;
	}

	LOG_INF("Publishing message: %s", log_strdup(CONFIG_CLOUD_MESSAGE));

	struct cloud_msg msg = {
		.qos = CLOUD_QOS_AT_MOST_ONCE,
		.buf = CONFIG_CLOUD_MESSAGE,
		.len = strlen(CONFIG_CLOUD_MESSAGE)
	};

	/* When using the nRF Cloud backend data is sent to the message topic.
	 * This is in order to visualize the data in the web UI terminal.
	 * For Azure IoT Hub and AWS IoT, messages are addressed directly to the
	 * device twin (Azure) or device shadow (AWS).
	 */
	if (strcmp(CONFIG_CLOUD_BACKEND, "NRF_CLOUD") == 0) {
		msg.endpoint.type = CLOUD_EP_MSG;
	} else {
		msg.endpoint.type = CLOUD_EP_STATE;
	}

	err = cloud_send(cloud_backend, &msg);
	if (err) {
		LOG_ERR("cloud_send failed, error: %d", err);
	}

#if defined(CONFIG_CLOUD_PUBLICATION_SEQUENTIAL)
	k_delayed_work_submit(
			&cloud_update_work,
			K_SECONDS(CONFIG_CLOUD_MESSAGE_PUBLICATION_INTERVAL));
#endif
}

void cloud_event_handler(const struct cloud_backend *const backend,
			 const struct cloud_event *const evt,
			 void *user_data)
{
	ARG_UNUSED(user_data);
	ARG_UNUSED(backend);

	switch (evt->type) {
	case CLOUD_EVT_CONNECTING:
		LOG_INF("CLOUD_EVT_CONNECTING");
		dk_set_led(DK_LED1, 0); //Turn on LED1 on connection to LTE network
		break;
	case CLOUD_EVT_CONNECTED:
		LOG_INF("CLOUD_EVT_CONNECTED");
		cloud_connected = true;
		break;
	case CLOUD_EVT_READY:
		LOG_INF("CLOUD_EVT_READY");
		dk_set_led(DK_LED2, 0); //Turn on LED2 on connection to cloud
		k_delayed_work_cancel(&connect_work);

#if defined(CONFIG_CLOUD_PUBLICATION_SEQUENTIAL)
		k_delayed_work_submit(&cloud_update_work, K_NO_WAIT);
#endif
		break;
	case CLOUD_EVT_DISCONNECTED:
		LOG_INF("CLOUD_EVT_DISCONNECTED");

		cloud_connected = false;

		if (k_delayed_work_pending(&connect_work)) {
			break;
		}

		k_delayed_work_submit(&connect_work, K_NO_WAIT);
		break;
	case CLOUD_EVT_ERROR:
		LOG_INF("CLOUD_EVT_ERROR");
		break;
	case CLOUD_EVT_DATA_SENT:
		LOG_INF("CLOUD_EVT_DATA_SENT");
		break;
	case CLOUD_EVT_DATA_RECEIVED:
		LOG_INF("CLOUD_EVT_DATA_RECEIVED");
		LOG_INF("Data received from cloud: %.*s",
			evt->data.msg.len,
			log_strdup(evt->data.msg.buf));
/*  -------------- JSON message received -------------  */
                {
                   int i;
                  
                   for (i=0; i<(evt->data.msg.len); i++)
                         rx_buffer[i] = evt->data.msg.buf[i];

                   rx_buffer[i] = 0;  // Add terminating 0
                   k_sem_give(&json_received);
                }
/*  --------------------------------------------------  */
		break;
	case CLOUD_EVT_PAIR_REQUEST:
		LOG_INF("CLOUD_EVT_PAIR_REQUEST");
		break;
	case CLOUD_EVT_PAIR_DONE:
		LOG_INF("CLOUD_EVT_PAIR_DONE");
		break;
	case CLOUD_EVT_FOTA_DONE:
		LOG_INF("CLOUD_EVT_FOTA_DONE");
		break;
	case CLOUD_EVT_FOTA_ERROR:
		LOG_INF("CLOUD_EVT_FOTA_ERROR");
		break;
	default:
		LOG_INF("Unknown cloud event type: %d", evt->type);
		break;
	}
}

static void work_init(void)
{
	k_delayed_work_init(&cloud_update_work, cloud_update_work_fn);
	k_delayed_work_init(&connect_work, connect_work_fn);
}

static void lte_handler(const struct lte_lc_evt *const evt)
{
	switch (evt->type) {
	case LTE_LC_EVT_NW_REG_STATUS:
		if ((evt->nw_reg_status != LTE_LC_NW_REG_REGISTERED_HOME) &&
		     (evt->nw_reg_status != LTE_LC_NW_REG_REGISTERED_ROAMING)) {
			break;
		}

		LOG_INF("Network registration status: %s",
			evt->nw_reg_status == LTE_LC_NW_REG_REGISTERED_HOME ?
			"Connected - home network" : "Connected - roaming");
		k_sem_give(&lte_connected);
		break;
	case LTE_LC_EVT_PSM_UPDATE:
		LOG_DBG("PSM parameter update: TAU: %d, Active time: %d",
			evt->psm_cfg.tau, evt->psm_cfg.active_time);
		break;
	case LTE_LC_EVT_EDRX_UPDATE: {
		char log_buf[60];
		ssize_t len;

		len = snprintf(log_buf, sizeof(log_buf),
			       "eDRX parameter update: eDRX: %f, PTW: %f",
			       evt->edrx_cfg.edrx, evt->edrx_cfg.ptw);
		if (len > 0) {
			LOG_DBG("%s", log_strdup(log_buf));
		}
		break;
	}
	case LTE_LC_EVT_RRC_UPDATE:
		LOG_DBG("RRC mode: %s",
			evt->rrc_mode == LTE_LC_RRC_MODE_CONNECTED ?
			"Connected" : "Idle");
		break;
	case LTE_LC_EVT_CELL_UPDATE:
		LOG_DBG("LTE cell changed: Cell ID: %d, Tracking area: %d",
			evt->cell.id, evt->cell.tac);
		break;
	default:
		break;
	}
}

static void modem_configure(void)
{
#if defined(CONFIG_NRF_MODEM_LIB)
	if (IS_ENABLED(CONFIG_LTE_AUTO_INIT_AND_CONNECT)) {
		/* Do nothing, modem is already configured and LTE connected. */
	} else {
		int err;

#if defined(CONFIG_POWER_SAVING_MODE_ENABLE)
		/* Requesting PSM before connecting allows the modem to inform
		 * the network about our wish for certain PSM configuration
		 * already in the connection procedure instead of in a separate
		 * request after the connection is in place, which may be
		 * rejected in some networks.
		 */
		err = lte_lc_psm_req(true);
		if (err) {
			LOG_ERR("Failed to set PSM parameters, error: %d",
				err);
		}

		LOG_INF("PSM mode requested");
#endif
		err = lte_lc_init_and_connect_async(lte_handler);
		if (err) {
			LOG_ERR("Modem could not be configured, error: %d",
				err);
			return;
		}

		/* Check LTE events of type LTE_LC_EVT_NW_REG_STATUS in
		 * lte_handler() to determine when the LTE link is up.
		 */
	}
#endif
}

#if defined(CONFIG_CLOUD_PUBLICATION_BUTTON_PRESS)
static void button_handler(uint32_t button_states, uint32_t has_changed)
{
	if (has_changed & button_states & DK_BTN1_MSK) {
		k_delayed_work_submit(&cloud_update_work, K_NO_WAIT);
	}
}
#endif
//========================================================
/*  -------------- Send data to nRF CLoud ------------  */
void SendTemperatureToCloud(void)
{
	int err;

	if (!cloud_connected) {
		LOG_INF("Not connected to cloud, abort cloud publication");
		return;
	}

	struct cloud_msg status_msg = {
		.qos = CLOUD_QOS_AT_MOST_ONCE,
		.buf = json_temp_buffer,
		.len = strlen(json_temp_buffer)
	};


	LOG_INF("Publishing json message: %s", log_strdup(json_temp_buffer));


	/* When using the nRF Cloud backend data is sent to the message topic.
	 * This is in order to visualize the data in the web UI terminal.
	 * For Azure IoT Hub and AWS IoT, messages are addressed directly to the
	 * device twin (Azure) or device shadow (AWS).
	 */
	if (strcmp(CONFIG_CLOUD_BACKEND, "NRF_CLOUD") == 0) {
		status_msg.endpoint.type = CLOUD_EP_MSG;
	} else {
		status_msg.endpoint.type = CLOUD_EP_STATE;
	}

	err = cloud_send(cloud_backend, &status_msg);
	if (err) {
		LOG_ERR("cloud_send failed, error: %d", err);
	}
}
//========================================================
/* Send json string with attribute status (LEDs etc.) to nRF Cloud: */

void SendAttributeStatusToCloud(void)
{
	int err;

	if (!cloud_connected) {
		LOG_INF("Not connected to cloud, abort cloud publication");
		return;
	}

	struct cloud_msg status_msg = {
		.qos = CLOUD_QOS_AT_MOST_ONCE,
		.buf = json_tx_buffer,
		.len = strlen(json_tx_buffer)
	};


	LOG_INF("Publishing json message: %s", log_strdup(json_tx_buffer));


	/* When using the nRF Cloud backend data is sent to the message topic.
	 * This is in order to visualize the data in the web UI terminal.
	 * For Azure IoT Hub and AWS IoT, messages are addressed directly to the
	 * device twin (Azure) or device shadow (AWS).
	 */
	if (strcmp(CONFIG_CLOUD_BACKEND, "NRF_CLOUD") == 0) {
		status_msg.endpoint.type = CLOUD_EP_MSG;
	} else {
		status_msg.endpoint.type = CLOUD_EP_STATE;
	}

	err = cloud_send(cloud_backend, &status_msg);
	if (err) {
		LOG_ERR("cloud_send failed, error: %d", err);
	}
}
/* --------------------------------------------------- */
/* -------------------- Main-loop -------------------- */
/* --------------------------------------------------- */

void main(void)
{
	int err;

        InitJsonCommandInterpreter();  // Initialize json command interpreter at startup

	LOG_INF("Cloud client has started");

	cloud_backend = cloud_get_binding(CONFIG_CLOUD_BACKEND);
	__ASSERT(cloud_backend != NULL, "%s backend not found",
		 CONFIG_CLOUD_BACKEND);

	err = cloud_init(cloud_backend, cloud_event_handler);
	if (err) {
		LOG_ERR("Cloud backend could not be initialized, error: %d",
			err);
	}

	work_init();
	modem_configure();

#if defined(CONFIG_CLOUD_PUBLICATION_BUTTON_PRESS)
	err = dk_buttons_init(button_handler);
	if (err) {
		LOG_ERR("dk_buttons_init, error: %d", err);
	}
#endif
	LOG_INF("Connecting to LTE network, this may take several minutes...");

	k_sem_take(&lte_connected, K_FOREVER);

	LOG_INF("Connected to LTE network");
	LOG_INF("Connecting to cloud");

	k_delayed_work_submit(&connect_work, K_NO_WAIT);

//========================================================
        while (1)  // Main-loop to handle received json command strings:
        {
              k_sem_take(&json_received, K_FOREVER);

              LOG_INF("JSON message received: %s\n", rx_buffer);
              ParseJsonString(rx_buffer);
              HandleUpdatedAttibuteValues();
              
              if (get_command_received)  // If a "GET" command is received ..
              {
                  SendAttributeStatusToCloud();  // the requested value will be send to cloud
                  get_command_received = FALSE;
              }
        }
//========================================================
}
